#include <algorithm>
#include <cstdlib>
#include <fstream>
#include <iostream>

#include <rlcsa/misc/utils.h>

#include "pdltree.h"


int
main(int argc, char** argv)
{
  std::cout << "Building PDL tree and sets" << std::endl;
  if(argc < 5)
  {
    std::cout << "Usage: build_pdltree -t base_name block_size max_set" << std::endl;
    std::cout << "  Builds a tree for PDLTopk." << std::endl;
    std::cout << "Usage: build_pdltree -r base_name block_size storing_factor" << std::endl;
    std::cout << "  Builds a tree for PDLRP (default)." << std::endl;
    std::cout << "Usage: build_pdltree -s base_name block_size storing_factor" << std::endl;
    std::cout << "  Builds a tree for PDL-set." << std::endl;
    return 1;
  }

  bool topk = (argc >= 5 && argv[1][0] == '-' && argv[1][1] == 't');
  bool use_rpset = (argc >= 5 && argv[1][0] == '-' && argv[1][1] == 's');
  std::cout << "Mode: ";
  if(topk) { std::cout << "Top-k"; }
  else if(use_rpset) { std::cout << "RP-set"; }
  else { std::cout << "RP"; }
  std::cout << std::endl;
  std::string base_name = argv[2];
  std::cout << "Base name: " << base_name << std::endl;
  usint block_size = atoi(argv[3]);
  std::cout << "Block size: " << block_size << std::endl;
  usint second_parameter = atoi(argv[4]);
  if(topk) { std::cout << "Maximum set size: " << second_parameter << std::endl; }
  else { std::cout << "Storing factor: " << second_parameter << std::endl; }
  std::cout << std::endl;

  RLCSA rlcsa(base_name);
  if(!(rlcsa.isOk())) { return 2; }
  rlcsa.printInfo();
  rlcsa.reportSize(true);
  usint size = rlcsa.getSize();

  double start = CSA::readTimer();
  usint flags = PDLTree::PRINT_FLAG;
  if(topk) { flags |= PDLTree::TOPK_FLAG; }
  else { flags |= PDLTree::CONTAINS_ALL_FLAG; }
  PDLTree pdltree(rlcsa, block_size, second_parameter, flags);
  if(!(pdltree.isOk())) { return 3; }
  double total = CSA::readTimer() - start;

  std::cout << "Writing tree..." << std::endl;
  std::string pdltree_name = base_name + PDLTree::EXTENSION;
  std::ofstream pdltree_file(pdltree_name.c_str(), std::ios_base::binary);
  if(!pdltree_file)
  {
    std::cerr << "build_pdltree: Cannot open output file " << pdltree_name << "!" << std::endl;
    return 4;
  }
  pdltree.writeTo(pdltree_file);

  usint freq_bytes = 0;
  if(topk)
  {
    std::cout << "Writing frequencies..." << std::endl;
    CSA::DeltaMultiArray* freqs = pdltree.getFrequencies(true);
    if(freqs != 0)
    {
      freq_bytes = freqs->reportSize();
      freqs->writeTo(pdltree_file);
      delete freqs; freqs = 0;
    }
  }
  pdltree_file.close();

  std::cout << "Writing sets..." << std::endl;
  std::string set_name = base_name + PDLTree::SETS_EXTENSION;
  std::ofstream set_file(set_name.c_str(), std::ios_base::binary);
  if(!set_file)
  {
    std::cerr << "build_pdltree: Cannot open output file " << set_name << "!" << std::endl;
    return 6;
  }
  pdltree.writeSets(set_file, (topk ? PDLTree::sort_topk : PDLTree::sort_id), use_rpset);
  set_file.close();
  std::cout << std::endl;

  double megabytes = size / (double)CSA::MEGABYTE;
  double pdl_size = pdltree.reportSize() / (double)CSA::MEGABYTE;
  double freq_size = freq_bytes / (double)CSA::MEGABYTE;
  double total_size = pdl_size + freq_size;
  std::cout << "Processed " << megabytes << " megabytes in " << total << " seconds (" << (megabytes / total) << " MB/s)." << std::endl;
  std::cout << "Memory: " << CSA::memoryUsage() << " kB" << std::endl;
  std::cout << "PDLTree: " << pdl_size << " MB (" << (CHAR_BIT * pdl_size / megabytes) << " bpc)" << std::endl;
  std::cout << "Freqs: " << freq_size << " MB (" << (CHAR_BIT * freq_size / megabytes) << " bpc)" << std::endl;
  std::cout << "Total: " << total_size << " MB (" << (CHAR_BIT * total_size / megabytes) << " bpc)" << std::endl;
  std::cout << std::endl;

  return 0;
}
