#include <algorithm>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <string>
#include <vector>

#include <rlcsa/alphabet.h>
#include <rlcsa/bits/deltavector.h>
#include <rlcsa/misc/utils.h>

//--------------------------------------------------------------------------

using CSA::usint;
using CSA::pair_type;

struct Document
{
  char* data;
  usint size;

  Document() : data(0), size(0) {}
  ~Document() { delete this->data; this->data = 0; }

  Document(std::ifstream& input, usint n) :
    data(0), size(n)
  {
    if(this->size > 0)
    {
      this->data = new char[this->size + 1];
      input.read(this->data, this->size);
    }
    else
    {
      std::string buffer;
      std::getline(input, buffer, '\0');
      this->size = buffer.length();
      this->data = new char[this->size + 1];
      buffer.copy(this->data, this->size);
    }
    this->data[this->size] = 0;
  }

  Document(const Document& original) :
    data(new char[original.size + 1]), size(original.size)
  {
    for(usint i = 0; i < this->size; i++) { this->data[i] = original.data[i]; }
    this->data[this->size] = 0;
  }

  Document(const Document& original, usint n) :
    data(new char[n * original.size + 1]), size(n * original.size)
  {
    for(usint copy = 0, i = 0; copy < n; copy++)
    {
      for(usint j = 0; j < original.size; i++, j++) { this->data[i] = original.data[j]; }
    }
    this->data[this->size] = 0;
  }

  Document& operator= (const Document& original)
  {
    if(this == &original) { return *this; }

    delete this->data;
    this->data = new char[original.size + 1]; this->size = original.size;
    for(usint i = 0; i < this->size; i++) { this->data[i] = original.data[i]; }
    this->data[this->size] = 0;

    return *this;
  }

  void writeTo(std::ofstream& output)
  {
    output.write(this->data, this->size + 1);
  }
};

typedef std::vector<Document> collection_type;

void printUsage();
void collectionSize(collection_type* collection);
void writeCollection(collection_type* collection, std::ofstream& output);

collection_type* concatenate(collection_type* collection, usint n); 
collection_type* duplicate(collection_type* collection, usint n); 
void mutate(collection_type* collection, usint* counts, double p);
void readDocuments(std::ifstream& input, collection_type* collection, usint* counts, usint num_docs, usint n);
usint setNumDocs(usint num_docs);

//--------------------------------------------------------------------------

int
main(int argc, char** argv)
{
  std::cout << "Synthetic document collection generator" << std::endl;
  if(argc < 3)
  {
    printUsage();
    return 1;
  }
  std::cout << "Input: " << argv[1] << std::endl;
  std::cout << "Output: " << argv[2] << std::endl;
  std::cout << std::endl;

  std::ifstream input(argv[1], std::ios_base::binary);
  if(!input)
  {
    std::cerr << "Mutate: Cannot open input file " << argv[1] << "!" << std::endl;
    return 2;
  }
  collection_type* collection = new collection_type;
  usint counts[CSA::CHARS];
  for(usint i = 0; i < CSA::CHARS; i++) { counts[i] = 0; }
  srand(0xDEADBEEF);  // FIXME random seed

  usint num_docs = 1;
  for(int cmd = 3; cmd < argc; cmd++)
  {
    usint as_int = atoi(argv[cmd] + 1);
    double as_float = atof(argv[cmd] + 1);
    switch(argv[cmd][0])
    {
      case 'c':
        collection = concatenate(collection, as_int);
        break;
      case 'd':
        collection = duplicate(collection, as_int);
        break;
      case 'm':
        mutate(collection, counts, as_float);
        break;
      case 'n':
        num_docs = setNumDocs(as_int);
        break;
      case 'r':
        readDocuments(input, collection, counts, num_docs, as_int);
        break;
      default:
        std::cerr << "Mutate: Invalid command: " << argv[cmd] << std::endl;
        break;
    }
    collectionSize(collection);
    std::cout << std::endl;
  }
  input.close();

  std::ofstream output(argv[2], std::ios_base::binary);
  if(!output)
  {
    std::cerr << "Mutate: Cannot open output file " << argv[2] << "!" << std::endl;
    return 3;
  }
  writeCollection(collection, output);
  output.close();

  delete collection; collection = 0;
  return 0;
}

//--------------------------------------------------------------------------

void
printUsage()
{
  std::cout << "Usage: mutator input output [commands]" << std::endl;
  std::cout << "  c#  Concatenate each document with itself # times" << std::endl;
  std::cout << "  d#  Duplicate each document # times" << std::endl;
  std::cout << "  m#  Apply point mutations to each text position with probability #" << std::endl;
  std::cout << "  n#  Set the number of documents to be read to # (default 1)" << std::endl;
  std::cout << "  r#  Read #-byte documents (0 means read until next end marker)" << std::endl;
  std::cout << std::endl;
}

void
collectionSize(collection_type* collection)
{
  if(collection == 0) { return; }

  usint size = 0;
  for(collection_type::iterator iter = collection->begin(); iter != collection->end(); ++iter)
  {
    size += iter->size;
  }
  double megabytes = size / (double)CSA::MEGABYTE;
  std::cout << "The collection contains " << collection->size();
  if(collection->size() == 1) { std::cout << " document of size "; }
  else { std::cout << " documents of total size "; }
  std::cout << megabytes << " MB." << std::endl;
}

void
writeCollection(collection_type* collection, std::ofstream& output)
{
  if(collection == 0) { return; }
  for(collection_type::iterator iter = collection->begin(); iter != collection->end(); ++iter)
  {
    iter->writeTo(output);
  }
}

//--------------------------------------------------------------------------

collection_type*
concatenate(collection_type* collection, usint n)
{
  if(collection == 0) { return 0; }

  std::cout << "Concatenate(" << n << ")" << std::endl;
  if(n == 0)
  {
    std::cerr << "Mutator::concatenate(): Cannot concatenate 0 times!" << std::endl;
    return collection;
  }
  if(n == 1) { return collection; }

  collection_type* result = new collection_type; result->reserve(n * collection->size());
  for(collection_type::iterator iter = collection->begin(); iter != collection->end(); ++iter)
  {
    result->push_back(Document(*iter, n));
  }

  delete collection; collection = 0;
  return result;
}

collection_type*
duplicate(collection_type* collection, usint n)
{
  if(collection == 0) { return 0; }

  std::cout << "Duplicate(" << n << ")" << std::endl;
  if(n == 0)
  {
    std::cerr << "Mutator::duplicate(): Cannot duplicate 0 times!" << std::endl;
    return collection;
  }
  if(n == 1) { return collection; }

  collection_type* result = new collection_type; result->reserve(n * collection->size());
  for(collection_type::iterator iter = collection->begin(); iter != collection->end(); ++iter)
  {
    for(usint i = 0; i < n; i++) { result->push_back(*iter); }
  }

  delete collection; collection = 0;
  return result;
}

void
mutate(collection_type* collection, usint* counts, double p)
{
  if(collection == 0 || counts == 0) { return; }

  std::cout << "Mutate(" << p << ")" << std::endl;
  if(collection->empty() || p <= 0.0) { return; }
  usint total = 0;
  for(usint i = 0; i < CSA::CHARS; i++) { total += counts[i]; }
  if(total == 0)
  {
    std::cerr << "Mutator::mutate(): Empty character distribution!" << std::endl;
    return;
  }

  CSA::Alphabet alphabet(counts);
  usint mutations = 0;
  for(collection_type::iterator iter = collection->begin(); iter != collection->end(); ++iter)
  {
    for(usint i = 0; i < iter->size; i++)
    {
      if(rand() / (RAND_MAX + 1.0) < p)
      {
        iter->data[i] = alphabet.charAt(rand() % total);
        mutations++;
      }
    }
  }
  std::cout << "There were " << mutations << " mutations." << std::endl;
}

void
readDocuments(std::ifstream& input, collection_type* collection, usint* counts, usint num_docs, usint n)
{
  if(collection == 0 || counts == 0) { return; }

  std::cout << "readDocuments(" << num_docs << ", " << n << ")" << std::endl;
  if(num_docs == 0) { return; }

  for(usint j = 0; j < num_docs; j++)
  {
    Document doc(input, n);
    for(usint i = 0; i < doc.size; i++)
    {
      if(doc.data[i] == 0) { std::cerr << "Mutator::readDocument(): The document contains a \\0!" << std::endl; }
      counts[(unsigned char)(doc.data[i])]++;
    }
    collection->push_back(doc);
  }
}

usint
setNumDocs(usint num_docs)
{
  std::cout << "setNumDocs(" << num_docs << ")" << std::endl;
  return num_docs;
}

//--------------------------------------------------------------------------
