#include <algorithm>
#include <fstream>
#include <iostream>
#include <stack>

#include "pdltopk.h"

//--------------------------------------------------------------------------

const std::string PDLTopk::EXTENSION = ".topk";

PDLTopk::PDLTopk(const RLCSA& _rlcsa, const std::string& base_name, bool compress_sets) :
  rlcsa(_rlcsa), status(st_error),
  tree(0), freqs(0), grammar(0), blocks(0)
{
  if(!(this->rlcsa.isOk()) || !(this->rlcsa.supportsLocate()))
  {
    std::cerr << "PDLTopk::PDLTopk(): Invalid RLCSA!" << std::endl;
    return;
  }

  std::string input_name = base_name + (compress_sets ? PDLTree::EXTENSION : EXTENSION);
  std::ifstream input(input_name.c_str(), std::ios_base::binary);
  if(!input)
  {
    std::cerr << "PDLTopk::PDLTopk(): Cannot open input file " << input_name << "!" << std::endl;
    return;
  }
  this->tree = new PDLTree(this->rlcsa, input);
  if(!(this->tree->isOk())) { return; }
  this->freqs = CSA::MultiArray::readFrom(input);
  if(!(this->freqs->isOk())) { return; }
  this->status = st_unfinished;

  if(compress_sets)
  {
    uint terminals = 0;
    this->grammar = readGrammar(*(this->tree), base_name, terminals);
    this->blocks = readBlocks(*(this->tree), base_name, terminals);
  }
  else
  {
    pair_type temp;
    input.read((char*)&temp, sizeof(temp)); // Items, bits.
    this->grammar = new CSA::ReadBuffer(input, temp.first, temp.second);
    this->blocks = CSA::MultiArray::readFrom(input);
  }
  input.close();
  if(this->blocks == 0 || !(this->blocks->isOk())) { return; }

  this->status = st_ok;
}

PDLTopk::~PDLTopk()
{
  delete this->tree; this->tree = 0;
  delete this->freqs; this->freqs = 0;
  delete this->grammar; this->grammar = 0;
  delete this->blocks; this->blocks = 0;
}

void
PDLTopk::writeTo(const std::string& base_name) const
{
  if(!(this->isOk()))
  {
    std::cerr << "PDLTopk::writeTo(): Status is not ok!" << std::endl;
    return;
  }

  std::string filename = base_name + EXTENSION;
  std::ofstream output(filename.c_str(), std::ios_base::binary);
  if(!output)
  {
    std::cerr << "PDLTopk::writeTo(): Cannot open output file " << filename << "!" << std::endl;
    return;
  }

  this->tree->writeTo(output);
  this->freqs->writeTo(output);

  pair_type temp(this->grammar->getNumberOfItems(), this->grammar->getItemSize());
  output.write((char*)&temp, sizeof(temp));
  this->grammar->writeTo(output);
  this->blocks->writeTo(output);

  output.close();
}

usint
PDLTopk::reportSize(bool include_freqs) const
{
  usint bytes = sizeof(*this) + this->tree->reportSize() + this->grammar->reportSize() + this->blocks->reportSize();
  if(include_freqs) { bytes += this->freqs->reportSize(); }
  return bytes;
}

//--------------------------------------------------------------------------

PDLTopk::result_type*
PDLTopk::query(const std::string& pattern, usint k, bool get_freqs) const
{
  if(!(this->isOk())) { return 0; }
  pair_type sa_range = this->rlcsa.count(pattern);
  if(CSA::isEmpty(sa_range)) { return 0; }
  return this->queryUnsafe(sa_range, k, get_freqs);
}

PDLTopk::result_type*
PDLTopk::query(pair_type sa_range, usint k, bool get_freqs) const
{
  if(!(this->isOk()) || CSA::isEmpty(sa_range) || sa_range.second >= this->rlcsa.getSize()) { return 0; }
  return this->queryUnsafe(sa_range, k, get_freqs);
}

PDLTopk::result_type*
PDLTopk::queryUnsafe(pair_type sa_range, usint k, bool get_freqs) const
{
  usint block = this->tree->getBlock(sa_range);
  if(block >= this->tree->getNumberOfNodes())
  {
    return bruteForceTopk(this->rlcsa, sa_range, k);
  }

  result_type* results = new result_type;

  // Decompress the block.
  CSA::MultiArray::Iterator* block_iter = this->blocks->getIterator();
  block_iter->goToItem(block, 0); block_iter->setEnd(block + 1, 0);
  std::stack<usint> buffer;
  while(results->size() < k && !(block_iter->atEnd()))
  {
    usint temp = block_iter->nextItem();
    buffer.push(temp);
    while(results->size() < k && !(buffer.empty()))
    {
      usint value = buffer.top(); buffer.pop();
      if(this->tree->isTerminal(value)) { results->push_back(Document(value, 0)); }
      else
      {
        temp = value;
        value = this->tree->toRule(value);
        buffer.push(this->grammar->readItemConst(2 * value + 1));
        buffer.push(this->grammar->readItemConst(2 * value));
      }
    }
  }
  delete block_iter; block_iter = 0;

  if(get_freqs)
  {
    CSA::MultiArray::Iterator* freq_iter = this->freqs->getIterator();
    freq_iter->goToItem(block, 0);
    usint freq = 0, run = 0;
    for(result_type::iterator iter = results->begin(); iter != results->end(); ++iter)
    {
      if(run == 0) { freq = freq_iter->nextItem(); run = freq_iter->nextItem(); }
      iter->second = freq; run--;
    }
    delete freq_iter; freq_iter = 0;
  }

  return results;
}

//--------------------------------------------------------------------------
