#include <algorithm>
#include <cstdlib>
#include <limits>

#include <rlcsa/bits/bitbuffer.h>

#include "pdltree.h"
#include "repair.h"


//--------------------------------------------------------------------------

void
addPairs(RePairEncoder::set_type& set, RePairEncoder::map_type& pairs, usint max_pairs)
{
  for(usint i = 0; i < set.size(); i++)
  {
    for(usint j = i + 1; j < set.size(); j++)
    {
      RePairEncoder::rule_type temp(set[i], set[j]);
      RePairEncoder::map_type::iterator iter = pairs.find(temp);
      if(pairs.find(temp) == pairs.end())
      {
        if(pairs.size() < max_pairs) { pairs[temp] = 1; }
      }
      else { (iter->second)++; }
    }
  }
}

RePairEncoder::RePairEncoder(set_of_sets& input, bool print, usint min_freq, usint max_pairs, usint max_set) :
  ok(false),
  sets(input.size()), items(0),
  alphabet_size(0),
  number_of_rules(0), rule_layers(0),
  rules(), stored_sets()
{
  this->getStatistics(input);
  if(this->sets == 0 || this->items == 0) { return; }
  if(max_pairs == 0) { max_pairs = std::numeric_limits<usint>::max(); }

  usint freq = std::numeric_limits<usint>::max();
  if(print) { std::cout << "Sets: " << input.size() << std::endl; }
  usint old_offset = 0, input_size = this->items;
  while(freq >= min_freq)
  {
    if(print)
    {
      std::cout << "Rule layer: " << this->rule_layers << std::endl;
      std::cout << "  Alphabet size: " << this->newAlphabetSize() << std::endl;
      std::cout << "  Input size: " << input_size << std::endl;
    }

    // Count pairs occurring in the sets.
    // If there can potentially be too many pairs, look at the sets in random order.
    // FIXME Should instead start the next pass at that position.
    map_type* pairs = new map_type;
    usint offset = 0;
    for(usint i = 0; i < input.size(); i++)
    {
      usint j = (old_offset + i) % input.size();
      if(input[j].size() <= max_set) { addPairs(input[j], *pairs, max_pairs); }
      if(max_pairs == 0 || pairs->size() < max_pairs) { offset = (j + 1) % input.size(); }
    }
    old_offset = offset;
    if(print)
    {
      std::cout << "  Map size: " << pairs->size() << std::endl;
      if(offset > 0) { std::cout << "  Next offset: " << old_offset << std::endl; }
    }

    // Find potential rules that are frequent enough.
    freq = 0;
    std::vector<GrammarRule>* potential_rules = new std::vector<GrammarRule>;
    for(map_type::iterator iter = pairs->begin(); iter != pairs->end(); ++iter)
    {
      freq = std::max(freq, iter->second);
      if(iter->second >= min_freq) { potential_rules->push_back(GrammarRule(0, iter->second, iter->first)); }
    }
    delete pairs; pairs = 0;
    GrammarRuleComparator gr_comparator;
    if(potential_rules->size() >= PARALLEL_SORT_MINIMUM)
    {
      CSA::parallelSort(potential_rules->begin(), potential_rules->end(), gr_comparator);
    }
    else
    {
      CSA::sequentialSort(potential_rules->begin(), potential_rules->end(), gr_comparator);
    }
    if(print) { std::cout << "  Potential rules: " << potential_rules->size() << std::endl; }

    // Create new rules.
    RuleLayer* new_layer = new RuleLayer;
    CSA::WriteBuffer visited(this->newAlphabetSize(), 1);
    for(std::vector<GrammarRule>::iterator iter = potential_rules->begin(); iter != potential_rules->end(); ++iter)
    {
      if(visited.isSet(iter->items.first) || visited.isSet(iter->items.second)) { continue; }
      visited.setBit(iter->items.first); visited.setBit(iter->items.second);
      iter->id = this->newAlphabetSize(); this->number_of_rules++;
      new_layer->rules.push_back(*iter);
    }
    delete potential_rules; potential_rules = 0;
    if(print) { std::cout << "  Selected rules: " << new_layer->rules.size() << std::endl; }

    // Store the new rules.
    if(!(new_layer->rules.empty()))
    {
      new_layer->item_to_rule.assign(this->newAlphabetSize(), new_layer->rules.size());
      for(std::vector<GrammarRule>::iterator iter = new_layer->rules.begin(); iter != new_layer->rules.end(); ++iter)
      {
        new_layer->item_to_rule[iter->items.first] = iter - new_layer->rules.begin();
        new_layer->item_to_rule[iter->items.second] = iter - new_layer->rules.begin();
      }
      this->rules.push_back(new_layer);
      this->rule_layers++;
    }

    // Apply the new rules.
    for(set_of_sets::iterator iter = input.begin(); iter != input.end(); ++iter)
    {
      input_size -= this->applyRules(this->rule_layers - 1, *iter);
    }
  }

  // Store the sets
  this->stored_sets.swap(input);
  this->sets = this->stored_sets.size();
  input.clear();

  this->ok = true;
}

RePairEncoder::~RePairEncoder()
{
  for(std::vector<RuleLayer*>::iterator iter = this->rules.begin(); iter != this->rules.end(); ++iter)
  {
    delete *iter; *iter = 0;
  }
}

void
RePairEncoder::writeTo(const std::string& base_name) const
{
  if(!(this->isOk())) { return; }

  std::string grammar_name = base_name + PDLTree::SETS_EXTENSION + ".R";
  std::ofstream grammar_file(grammar_name.c_str(), std::ios_base::binary);
  if(!grammar_file)
  {
    std::cerr << "RePairEncoder::writeTo(): Cannot open output file " << grammar_name << "!" << std::endl;
    return;
  }

  uint temp = this->alphabet_size + this->sets; // Number of terminals.
  grammar_file.write((char*)&temp, sizeof(temp));
  for(usint i = 0; i < this->rule_layers; i++)
  {
    RuleLayer* layer = this->rules[i];
    for(std::vector<GrammarRule>::iterator iter = layer->rules.begin(); iter != layer->rules.end(); ++iter)
    {
      temp = iter->items.first; if(temp >= this->alphabet_size) { temp += this->sets; }
      grammar_file.write((char*)&temp, sizeof(temp));
      temp = iter->items.second; if(temp >= this->alphabet_size) { temp += this->sets; }
      grammar_file.write((char*)&temp, sizeof(temp));
    }
  }
  grammar_file.close();

  std::string blocks_name = base_name + PDLTree::SETS_EXTENSION + ".C";
  std::ofstream blocks_file(blocks_name.c_str(), std::ios_base::binary);
  if(!blocks_file)
  {
    std::cerr << "RePairEncoder::writeTo(): Cannot open output file " << blocks_name << "!" << std::endl;
    return;
  }

  for(usint i = 0; i < this->sets; i++)
  {
    set_type current = this->stored_sets[i];
    for(set_type::iterator iter = current.begin(); iter != current.end(); ++iter)
    {
      temp = *iter; if(temp >= this->alphabet_size) { temp += this->sets; }
      blocks_file.write((char*)&temp, sizeof(temp));
    }
    temp = this->alphabet_size + i; // endmarker
    blocks_file.write((char*)&temp, sizeof(temp));
  }
  blocks_file.close();
}

//--------------------------------------------------------------------------

bool
RePairEncoder::addSet(set_type& new_set)
{
  if(!(this->isOk()) || new_set.empty()) { return false; }
  if(*std::max_element(new_set.begin(), new_set.end()) >= this->alphabet_size) { return false; }

  this->sets++; this->items += new_set.size();
  for(usint i = 0; i < this->rule_layers; i++) { this->applyRules(i, new_set); }
  this->stored_sets.push_back(new_set);

  return true;
}

//--------------------------------------------------------------------------

void
RePairEncoder::getStatistics(const set_of_sets& sets)
{
  for(set_of_sets::const_iterator iter = sets.begin(); iter != sets.end(); ++iter)
  {
    this->items += iter->size();
    if(!(iter->empty()))
    {
      this->alphabet_size = std::max(this->alphabet_size, (usint)(*std::max_element(iter->begin(), iter->end()) + 1));
    }
  }
}

bool
RePairEncoder::pairsCanExceed(usint limit) const
{
  usint estimate = this->newAlphabetSize();
  if(estimate == 0) { return false; }
  return (limit / estimate <= estimate / 2);  // A bit inexact.
}

usint
RePairEncoder::applyRules(usint layer, set_type& target) const
{
  if(layer >= this->rule_layers || target.empty()) { return 0; }

  // Find rules that cover each of the elements.
  RuleLayer* current_rules = this->rules[layer];
  set_type applicable_rules;
  for(set_type::iterator iter = target.begin(); iter != target.end(); ++iter)
  {
    usint pos = current_rules->item_to_rule[*iter];
    if(pos < current_rules->rules.size()) { applicable_rules.push_back(pos); }
  }
  if(applicable_rules.empty()) { return 0; }
  CSA::sequentialSort(applicable_rules.begin(), applicable_rules.end());

  // If a position occurs twice in applicable_rules, both items are covered by that rule.
  set_type items_to_remove;
  for(usint i = 1, prev = applicable_rules[0]; i < applicable_rules.size(); i++)
  {
    if(applicable_rules[i] == prev)
    {
      GrammarRule& rule = current_rules->rules[prev];
      target.push_back(rule.id);
      items_to_remove.push_back(rule.items.first);
      items_to_remove.push_back(rule.items.second);
    }
    prev = applicable_rules[i];
  }
  CSA::sequentialSort(items_to_remove.begin(), items_to_remove.end());

  // Remove items covered by applied rules.
  set_type::iterator set_iter = target.begin(), remove_iter = items_to_remove.begin();
  set_type new_set;
  while(set_iter != target.end() && remove_iter != items_to_remove.end())
  {
    if(*set_iter < *remove_iter) { new_set.push_back(*set_iter); ++set_iter; }
    else if(*set_iter > *remove_iter) { ++remove_iter; }
    else { ++set_iter; ++remove_iter; }
  }
  while(set_iter != target.end()) { new_set.push_back(*set_iter); ++set_iter; }

  target.swap(new_set);

  return items_to_remove.size() / 2;
}

//--------------------------------------------------------------------------
