#include <iostream>

#include <rlcsa/misc/utils.h>

#include "utils.h"

//--------------------------------------------------------------------------

CSA::ReadBuffer*
readDocarray(const RLCSA& rlcsa, const std::string& base_name)
{
  const static std::string DOCARRAY_EXTENSION = ".docs";

  std::string docarray_name = base_name + DOCARRAY_EXTENSION;
  std::ifstream input(docarray_name.c_str(), std::ios_base::binary);
  if(!input)
  {
    std::cerr << "readDocarray(): Cannot open input file " << docarray_name << "!" << std::endl;
    return 0;
  }

  CSA::ReadBuffer* docarray = readDocarray(rlcsa, input);
  input.close();
  return docarray;
}

CSA::ReadBuffer*
readDocarray(const RLCSA& rlcsa, std::ifstream& input)
{
  usint item_bits = std::max((usint)1, CSA::length(rlcsa.getNumberOfSequences() - 1));
  return new CSA::ReadBuffer(input, rlcsa.getSize(), item_bits);
}

//--------------------------------------------------------------------------

void
printResults(const std::string& algorithm, double seconds, usint docc, usint patterns)
{
  std::cout << algorithm << ':';
  for(usint i = algorithm.length() + 1; i < 14; i++) { std::cout << ' '; }
  std::cout << docc << " documents in " << seconds << " seconds ("
            << (docc / seconds) << " docc/s, "
            << (patterns / seconds) << " patterns/s)" << std::endl;
}

void
printSize(const std::string& algorithm, usint text_size, usint index_size, usint doclist_size)
{
  std::cout << algorithm << ':';
  for(usint i = algorithm.length() + 1; i < 14; i++) { std::cout << ' '; }

  double doclist_mb = doclist_size / (double)CSA::MEGABYTE;
  double total_mb = (index_size + doclist_size) / (double)CSA::MEGABYTE;
  double doclist_bpc = CHAR_BIT * doclist_size / (double)text_size;
  double total_bpc = CHAR_BIT * (index_size + doclist_size) / (double)text_size;

  std::cout << doclist_mb << " MB / " << doclist_bpc << " bpc (total "
            << total_mb << " MB / " << total_bpc << " bpc)" << std::endl;
}

void
printSizeNoCSA(const std::string& algorithm, usint text_size, usint doclist_size)
{
  std::cout << algorithm << ':';
  for(usint i = algorithm.length() + 1; i < 14; i++) { std::cout << ' '; }

  double doclist_mb = doclist_size / (double)CSA::MEGABYTE;
  double doclist_bpc = CHAR_BIT * doclist_size / (double)text_size;

  std::cout << doclist_mb << " MB / " << doclist_bpc << " bpc" << std::endl;
}

//--------------------------------------------------------------------------

void
sortById(std::vector<Document>* docs, bool parallel)
{
  if(parallel) { CSA::parallelSort(docs->begin(), docs->end()); }
  else { CSA::sequentialSort(docs->begin(), docs->end()); }
}

struct MostFrequentComparator
{
  bool operator() (const Document& a, const Document& b)
  {
    // Descending by frequency, ascending by id.
    if(a.second != b.second) { return (a.second > b.second); }
    return (a.first < b.first);
  }
} mf_comparator;

void
sortByFrequency(std::vector<Document>* docs, bool parallel)
{
  if(parallel) { CSA::parallelSort(docs->begin(), docs->end(), mf_comparator); }
  else { CSA::sequentialSort(docs->begin(), docs->end(), mf_comparator); }
}

void
mergeDocumentRuns(std::vector<Document>* docs)
{
  if(docs->empty()) { return; }

  std::vector<Document> another;
  Document curr(0, 0);
  for(std::vector<Document>::iterator iter = docs->begin(); iter != docs->end(); ++iter)
  {
    if(iter->first == curr.first) { curr.second += iter->second; }
    else
    {
      if(curr.second > 0) { another.push_back(curr); curr.second = 0; }
      curr = *iter;
    }
  }
  if(curr.second > 0) { another.push_back(curr); }

  docs->swap(another);
}

//--------------------------------------------------------------------------

std::vector<uint>*
bruteForceDoclistUnsafe(const RLCSA* rlcsa, const CSA::ReadBuffer* docarray, pair_type sa_range, std::vector<uint>* result)
{
  bool filter = (result == 0);
  if(result == 0) { result = new std::vector<uint>; result->reserve(CSA::length(sa_range)); }

  if(docarray == 0)
  {
    for(usint offset = sa_range.first; offset <= sa_range.second; offset += CSA::MILLION)
    {
      usint len = std::min(CSA::MILLION, sa_range.second + 1 - offset);
      usint* buffer = rlcsa->getSequenceForPosition(rlcsa->locate(sa_range), len);
      for(usint i = 0; i < len; i++) { result->push_back(buffer[i]); }
      delete[] buffer; buffer = 0;
    }
  }
  else
  {
    for(usint i = sa_range.first; i <= sa_range.second; i++)
    {
      result->push_back(docarray->readItemConst(i));
    }
  }

  if(filter) { CSA::removeDuplicates(result, false); }

  return result;
}

std::vector<uint>*
bruteForceDoclist(const RLCSA& rlcsa, const std::string& pattern, std::vector<uint>* result)
{
  if(!(rlcsa.isOk())) { return result; }
  pair_type sa_range = rlcsa.count(pattern);
  if(CSA::isEmpty(sa_range)) { return result; }
  return bruteForceDoclistUnsafe(&rlcsa, 0, sa_range, result);
}

std::vector<uint>*
bruteForceDoclist(const RLCSA& rlcsa, pair_type sa_range, std::vector<uint>* result)
{
  if(!(rlcsa.isOk()) || CSA::isEmpty(sa_range) || sa_range.second >= rlcsa.getSize()) { return result; }
  return bruteForceDoclistUnsafe(&rlcsa, 0, sa_range, result);
}

std::vector<uint>*
bruteForceDoclist(const RLCSA& rlcsa, const CSA::ReadBuffer& docarray, const std::string& pattern, std::vector<uint>* result)
{
  if(!(rlcsa.isOk())) { return result; }
  pair_type sa_range = rlcsa.count(pattern);
  if(CSA::isEmpty(sa_range)) { return result; }
  return bruteForceDoclistUnsafe(0, &docarray, sa_range, result);
}

std::vector<uint>*
bruteForceDoclist(const CSA::ReadBuffer& docarray, pair_type sa_range, std::vector<uint>* result)
{
  if(CSA::isEmpty(sa_range) || sa_range.second >= docarray.getNumberOfItems()) { return result; }
  return bruteForceDoclistUnsafe(0, &docarray, sa_range, result);
}

//--------------------------------------------------------------------------

std::vector<Document>*
bruteForceTopkUnsafe(const RLCSA* rlcsa, const CSA::ReadBuffer* docarray, pair_type sa_range, usint k)
{
  std::vector<Document>* results = new std::vector<Document>;
  results->reserve(CSA::length(sa_range));

  if(docarray == 0)
  {
    for(usint offset = sa_range.first; offset <= sa_range.second; offset += CSA::MILLION)
    {
      usint len = std::min(CSA::MILLION, sa_range.second + 1 - offset);
      usint* buffer = rlcsa->getSequenceForPosition(rlcsa->locate(sa_range), len);
      for(usint i = 0; i < len; i++) { results->push_back(Document(buffer[i], 1)); }
      delete[] buffer; buffer = 0;
    }
  }
  else
  {
    for(usint i = sa_range.first; i <= sa_range.second; i++)
    {
      results->push_back(Document(docarray->readItemConst(i), 1));
    }
  }

  sortById(results, false);
  mergeDocumentRuns(results);
  sortByFrequency(results, false);
  if(results->size() > k) { results->resize(k); }

  return results;
}

std::vector<Document>*
bruteForceTopk(const RLCSA& rlcsa, const std::string& pattern, usint k)
{
  if(!(rlcsa.isOk())) { return 0; }
  pair_type sa_range = rlcsa.count(pattern);
  if(CSA::isEmpty(sa_range)) { return 0; }
  return bruteForceTopkUnsafe(&rlcsa, 0, sa_range, k);
}

std::vector<Document>*
bruteForceTopk(const RLCSA& rlcsa, pair_type sa_range, usint k)
{
  if(!(rlcsa.isOk()) || CSA::isEmpty(sa_range) || sa_range.second >= rlcsa.getSize()) { return 0; }
  return bruteForceTopkUnsafe(&rlcsa, 0, sa_range, k);
}

std::vector<Document>*
bruteForceTopk(const RLCSA& rlcsa, const CSA::ReadBuffer& docarray, const std::string& pattern, usint k)
{
  if(!(rlcsa.isOk())) { return 0; }
  pair_type sa_range = rlcsa.count(pattern);
  if(CSA::isEmpty(sa_range)) { return 0; }
  return bruteForceTopkUnsafe(0, &docarray, sa_range, k);
}

std::vector<Document>*
bruteForceTopk(const CSA::ReadBuffer& docarray, pair_type sa_range, usint k)
{
  if(CSA::isEmpty(sa_range) || sa_range.second >= docarray.getNumberOfItems()) { return 0; }
  return bruteForceTopkUnsafe(0, &docarray, sa_range, k);
}

//--------------------------------------------------------------------------
